"""
Create draft intermediate mapping in CSV from source dataset to target dataset
"""

from __future__ import annotations

import argparse
import warnings
from pathlib import Path
from typing import Literal

import numpy as np
import pandas as pd

from .util import (
    DEFAULT_CONFIG,
    check_matches,
    load_data_dict,
    read_config_schema,
    read_json,
    setup_llm,
)


class Mapper:
    """
    Class for creating an intermediate mapping file linking the data dictionary to
    schema fields and values.

    Use `create_mapping()` to write out the mapping file, as the function equivalent
    of the command line `create-mapping` script.

    Parameters
    ----------
    data_dictionary
        The data dictionary to use
    schema
        The path to the schema file to map to
    language
        The language of the raw data (e.g. 'fr', 'en', 'es')
    api_key
        The API key to use for the LLM
    llm
        The LLM to use, currently only 'openai' and 'gemini' are supported
    config
        The path to the configuration file to use if not using the default configuration
    """

    def __init__(
        self,
        data_dictionary: str | pd.DataFrame,
        schema: Path,
        language: str,
        api_key: str | None = None,
        llm: Literal["openai", "gemini"] | None = "openai",
        config: Path | None = None,
    ):
        self.schema = read_json(schema)
        self.schema_properties = self.schema["properties"]
        self.language = language
        if llm is None:
            self.model = None
        else:
            self.model = setup_llm(llm, api_key)

        self.config = read_config_schema(
            config or Path(Path(__file__).parent, DEFAULT_CONFIG)
        )

        self.data_dictionary = load_data_dict(self.config, data_dictionary)

    @property
    def target_fields(self) -> list[str]:
        """Returns a list of fields in the target schema"""
        try:
            return self._target_fields
        except AttributeError:
            self._target_fields = list(self.schema_properties.keys())
            return self._target_fields

    @property
    def target_types(self) -> dict[str, list[str]]:
        """Returns the field types of the target schema"""
        try:
            return self._target_types
        except AttributeError:
            self._target_types = {
                f: self.schema_properties[f].get("type", ["string", "null"])
                for f in self.target_fields
            }
            return self._target_types

    @property
    def target_values(self) -> pd.Series:
        """Returns the enum values or boolean options for the target schema"""
        try:
            return self._target_values
        except AttributeError:

            def _value_options(f):
                if "boolean" in self.target_types[f]:
                    return ["True", "False", "None"]
                elif "string" in self.target_types[f]:
                    return self.schema_properties[f].get("enum", np.nan)
                elif "array" in self.target_types[f]:
                    return (
                        self.schema_properties[f].get("items", {}).get("enum", np.nan)
                    )
                else:
                    return np.nan

            self._target_values = pd.Series(
                {f: _value_options(f) for f in self.target_fields}, self.target_fields
            )

            return self._target_values

    @property
    def common_values(self) -> pd.Series:
        """
        Returns the commonly repeated values in the source data
        Usually this indicates that the source field is an enum or boolean
        """
        try:
            return self._common_values
        except AttributeError:
            if "common_values" in self.data_dictionary.columns:

                def _lower_string(x):
                    if isinstance(x, str):
                        return {
                            y.lower().strip()
                            for y in x.split(self.config["choice_delimiter"])
                        }

                cv = self.data_dictionary.common_values
                cv = cv.apply(_lower_string)
                cv.index = self.data_dictionary.source_field
                self._common_values = cv
            elif "choices" in self.data_dictionary.columns:
                raise NotImplementedError("choices column not yet supported")
            else:
                raise ValueError(
                    "No common values or choices column found in data dictionary"
                )
            return self._common_values

    @property
    def common_values_mapped(self) -> pd.Series:
        try:
            return self._common_values_mapped
        except AttributeError:
            try:
                filtered_dict = self.filtered_data_dict
            except AttributeError:
                raise AttributeError(
                    "fields have to be mapped using the `match_fields_to_schema` method"
                    " first"
                )
            cv = self.common_values
            # cv_mapped = filtered_dict.source_field.dropna().apply(lambda x: cv.loc[x])
            cv_mapped = filtered_dict.source_field.apply(
                lambda x: cv.loc[x] if isinstance(x, str) else None
            )
            self._common_values_mapped = cv_mapped
            return self._common_values_mapped

    @property
    def mapped_fields(self):
        try:
            return self._mapped_fields
        except AttributeError:
            raise AttributeError(
                "mapped_fields have to be created using `match_fields_to_schema` method"
            )

    @mapped_fields.setter
    def mapped_fields(self, value: pd.Series):
        self._mapped_fields = value

    def match_fields_to_schema(self) -> pd.DataFrame:
        """
        Use the LLM to match the target (schema) fields to the descriptions of the
        source data fields from the data dictionary.
        """

        # english translated descriptions rather than names.
        source_fields = list(self.data_dictionary.source_description)

        mappings = self.model.map_fields(source_fields, self.target_fields)

        mapping_dict = pd.DataFrame(
            {
                "target_field": [f.target_field for f in mappings.targets_descriptions],
                "source_description": [
                    f.source_description for f in mappings.targets_descriptions
                ],
            }
        )

        df_merged = pd.merge(
            mapping_dict,
            self.data_dictionary,
            how="left",
            on="source_description",
        )
        df_merged.set_index("target_field", inplace=True, drop=True)

        # Check to see if any fields with mapped descriptions are missing after merge
        missed_merge = df_merged[
            (df_merged["source_description"].notna())
            & (df_merged["source_field"].isna())
        ]

        if not missed_merge.empty:
            descriptions_list = self.data_dictionary["source_description"].tolist()
            df_merged.loc[
                (df_merged["source_description"].notna())
                & (df_merged["source_field"].isna()),
                "source_description",
            ] = missed_merge["source_description"].apply(
                lambda x: check_matches(x, descriptions_list)
            )

            df_merged = (
                df_merged["source_description"]
                .reset_index()
                .merge(self.data_dictionary, how="left")
                .set_index("target_field")
            )

        self.mapped_fields = df_merged.source_field
        self.filtered_data_dict = df_merged
        return df_merged

    def match_values_to_schema(self) -> pd.DataFrame:
        """
        Use the LLM to match the common values from the data dictionary to the target
        values in the schema - i.e. enum or boolean options.
        """

        values_tuples = []
        for f in self.target_fields:
            s = self.common_values_mapped.get(f)
            t = self.target_values[f]
            if s and t:
                values_tuples.append((f, s, t))

        # to LLM
        value_pairs = self.model.map_values(values_tuples, self.language)

        value_mapping = {}

        for p in value_pairs.values:
            f = p.field_name
            value_dict = {
                pair.source_value: pair.target_value for pair in p.mapped_values
            }
            value_mapping[f] = value_dict

        self.mapped_values = pd.Series(value_mapping, name="mapped_values")

        return self.mapped_values

    def create_mapping(self, save=True, file_name="mapping_file") -> pd.DataFrame:
        """
        Creates an intermediate mapping dataframe linking the data dictionary to schema
        fields. The index contains the target (schema) field names, and the columns are:
        source_description
        source_field
        common_values OR choices (depending on the data dictionary)
        target_values
        value_mapping

        Raises a warning if any fields are present in the schema where a
        corresponding source field in the data dictionary has not been found.

        Parameters
        ----------
        save
            Whether to save the mapping to a CSV file. If True, lists in `target_values`
            dicts in `value_mapping` are converted to strings before saving.
        name
            The name to use for the CSV file
        """

        mapping_dict = self.match_fields_to_schema()
        mapped_vals = self.match_values_to_schema()

        mapping_dict.drop(columns=["source_type"], inplace=True)

        # reindex to add in any schema fields that weren't returned by the LLM
        mapping_dict = mapping_dict.reindex(self.target_fields)

        mapping_dict["target_values"] = mapping_dict.index.map(self.target_values)
        mapping_dict["value_mapping"] = mapping_dict.index.map(mapped_vals)

        unmapped = mapping_dict[mapping_dict["source_field"].isna()].index
        if any(unmapped):
            warnings.warn(
                f"The following schema fields have not been mapped: {list(unmapped)}",
                UserWarning,
            )

        # turn lists & dicts into strings for consistancy with saved CSV
        mapping_dict["target_values"] = mapping_dict["target_values"].apply(
            lambda x: (", ".join(str(item) for item in x) if isinstance(x, list) else x)
        )
        mapping_dict["value_mapping"] = mapping_dict["value_mapping"].apply(
            lambda x: (
                ", ".join(f"{k}={v}" for k, v in x.items())
                if isinstance(x, dict)
                else x
            )
        )

        if save is False:
            return mapping_dict
        else:
            # Write to CSV
            if not file_name.endswith(".csv"):
                file_name += ".csv"
            mapping_dict.to_csv(file_name)
            return mapping_dict


def create_mapping(
    data_dictionary: str | pd.DataFrame,
    schema: Path,
    language: str,
    api_key: str,
    llm: str | None = "openai",
    config: Path | None = None,
    save: bool = True,
    file_name: str = "mapping_file",
) -> pd.DataFrame:
    """
    Creates a csv containing the mapping between a data dictionary and a schema.

    Takes a data dictionary and matches both the source fields, and any common values
    to the schema. Uses an LLM to first match the source fields to appropriate schema
    targets, and then to match the common values to appropriate enum or boolean options.

    Parameters
    ----------
    data_dictionary
        Path to a CSV or XLSX file, or a DataFrame, containing the data dictionary.
    schema
        Path to a JSON schema file.
    language
        Language of the source data (e.g. french, english, spanish).
    api_key
        API key for the API defined in `llm`
    llm
        Which LLM to use, currently only 'openai' is supported.
    config
        Path to a JSON file containing the configuration for autoparser.
    save
        Whether to save the mapping to a CSV file.
    file_name
        Name of the file to save the mapping to.

    Returns
    -------
    pd.DataFrame
        Dataframe containing the mapping between the data dictionary and the schema.
    """
    df = Mapper(data_dictionary, schema, language, api_key, llm, config).create_mapping(
        save=save, file_name=file_name
    )

    return df


def main(argv=None):
    parser = argparse.ArgumentParser(
        description=(
            "Generate intermediate CSV used by make_toml.py (create-parser) "
            "to create TOML"
        ),
        prog="autoparser create-mapping",
    )
    parser.add_argument("dictionary", help="Data dictionary to use")
    parser.add_argument("schema", help="Schema to use")
    parser.add_argument("language", help="Language of the original data")
    parser.add_argument("api_key", help="OpenAI API key to use")
    parser.add_argument("-l", "--llm", help="LLM API to use", default="openai")
    parser.add_argument(
        "-c",
        "--config",
        help=f"Configuration file to use (default={DEFAULT_CONFIG})",
        type=Path,
    )
    parser.add_argument(
        "-o", "--output", help="Name to use for output files", default="mapping_file"
    )
    args = parser.parse_args(argv)
    Mapper(
        args.dictionary,
        Path(args.schema),
        args.language,
        args.api_key,
        args.llm,
        args.config,
    ).create_mapping(save=True, file_name=args.output)


if __name__ == "__main__":
    main()



from __future__ import annotations

import copy
import csv
import io
import itertools
import json
import logging
import re
from collections import Counter, defaultdict
from functools import lru_cache
from pathlib import Path
from typing import Any, Iterable, Literal, Union

import fastjsonschema
import requests
import tomli
from more_itertools import unique_everseen
from tqdm.autonotebook import tqdm

from adtl.get_value import get_value, parse_if

SUPPORTED_FORMATS = {"json": json.load, "toml": tomli.load}
DEFAULT_DATE_FORMAT = "%Y-%m-%d"

StrDict = dict[str, Any]


def expand_refs(spec_fragment: StrDict, defs: StrDict) -> Union[StrDict, list[StrDict]]:
    "Expand all references (ref) with definitions (defs)"

    if spec_fragment == {}:
        return {}
    if isinstance(spec_fragment, dict):
        if "ref" in spec_fragment:
            reference_expanded = defs[spec_fragment["ref"]]
            del spec_fragment["ref"]
            spec_fragment = {**reference_expanded, **spec_fragment}
        return {k: expand_refs(spec_fragment[k], defs) for k in spec_fragment}
    elif isinstance(spec_fragment, list):
        return [expand_refs(m, defs) for m in spec_fragment]
    else:
        return spec_fragment


def expand_for(spec: list[StrDict]) -> list[StrDict]:
    "Expands for expressions in oneToMany table blocks"

    out = []

    def replace_val(
        item: Union[str, float, dict[str, Any]], replace: dict[str, Any]
    ) -> dict[str, Any]:
        block = {}
        if isinstance(item, str):
            return item.format(**replace)
        elif isinstance(item, (float, int)):
            return item
        for k, v in item.items():
            if not isinstance(k, str):
                block[k] = v
                continue
            rk = k.format(**replace)
            if isinstance(v, dict):
                block[rk] = replace_val(v, replace)
            elif isinstance(v, str):
                block[rk] = v.format(**replace)
            elif isinstance(v, list):
                block[rk] = [replace_val(it, replace) for it in v]
            else:
                block[rk] = v
        return block

    for match in spec:
        if "for" not in match:
            out.append(match)
            continue
        for_expr = match.pop("for")
        if not isinstance(for_expr, dict):
            raise ValueError(
                f"for expression {for_expr!r} is not a dictionary of variables to list "
                "of values or a range"
            )

        # Expand ranges when available
        for var in for_expr:
            if (
                "range" in for_expr[var]
                and isinstance(for_expr[var]["range"], list)
                and len(for_expr[var]["range"]) == 2
                and isinstance(for_expr[var]["range"][0], int)
                and isinstance(for_expr[var]["range"][1], int)
                and for_expr[var]["range"][1] > for_expr[var]["range"][0]
            ):
                start, end = for_expr[var]["range"]
                for_expr[var] = range(start, end + 1)  # add one to include end in list
            elif isinstance(for_expr[var], list):
                pass
            else:
                raise ValueError(
                    f"for expression {for_expr!r} can only have lists or ranges for "
                    "variables"
                )
        loop_vars = sorted(for_expr.keys())
        loop_assignments = [
            dict(zip(loop_vars, vals))
            for vals in itertools.product(*(for_expr[var] for var in loop_vars))
        ]
        for replacement in loop_assignments:
            out.append(replace_val(match, replacement))
    return out


def remove_null_keys(d: dict[str, Any]) -> dict[str, Any]:
    "Removes keys which map to null - but not empty strings or 'unknown' etc types"
    return {k: v for k, v in d.items() if v is not None}


def get_date_fields(schema: dict[str, Any]) -> list[str]:
    "Returns list of date fields from schema"
    fields = [
        field
        for field in schema["properties"]
        if field == "date" or "date_" in field or "_date" in field
    ]
    format_date_fields = [
        field
        for field in schema["properties"]
        if schema["properties"][field].get("format") == "date"
    ]
    return sorted(set(fields + format_date_fields))


def make_fields_optional(
    schema: dict[str, Any], optional_fields: list[str]
) -> dict[str, Any]:
    "Returns JSON schema with required fields modified to drop optional fields"
    if optional_fields is None:
        return schema
    _schema = copy.deepcopy(schema)
    _schema["required"] = sorted(set(schema["required"]) - set(optional_fields))
    for opt in ["oneOf", "anyOf"]:
        if opt in _schema:
            if any("required" in _schema[opt][x] for x in range(len(_schema[opt]))):
                for x in range(len(_schema[opt])):
                    _schema[opt][x]["required"] = list(
                        set(_schema[opt][x]["required"]) - set(optional_fields or [])
                    )
                if all(
                    all(bool(v) is False for v in _schema[opt][x].values())
                    for x in range(len(_schema[opt]))
                ):
                    _schema.pop(opt)
                else:
                    _schema[opt] = list(unique_everseen(_schema[opt]))
    return _schema


def relative_path(source_file, target_file):
    return Path(source_file).parent / target_file


def read_definition(file: Path) -> dict[str, Any]:
    "Reads definition from file into a dictionary"
    if isinstance(file, str):
        file = Path(file)
    if file.suffix == ".json":
        with file.open() as fp:
            return json.load(fp)
    elif file.suffix == ".toml":
        with file.open("rb") as fp:
            return tomli.load(fp)
    else:
        raise ValueError(f"Unsupported file format: {file}")


class Parser:
    """Main parser class that loads a specification

    Typical use of this within Python code::

        import adtl

        parser = adtl.Parser(specification)
        print(parser.tables) # list of tables created

        for row in parser.parse().read_table(table):
            print(row)
    """

    def __init__(
        self,
        spec: Union[str, Path, StrDict],
        include_defs: list[str] = [],
        quiet: bool = False,
    ):
        """Loads specification from spec in format (default json)

        Args:
            spec: Either the specification file to read (as Path or str), or
                the specification loaded into a dictionary
            include_defs: Definition files to include. These are spliced
                directly into the adtl.defs section of the :ref:`specification`.
            quiet: Boolean that switches on the verbosity of the parser, default False
        """

        self.data: StrDict = {}
        self.defs: StrDict = {}
        self.fieldnames: dict[str, list[str]] = {}
        self.specfile = None
        self.include_defs = include_defs
        self.validators: StrDict = {}
        self.schemas: StrDict = {}
        self.quiet = quiet
        self.date_fields = []
        self.report = {
            "validation_errors": defaultdict(Counter),
            "total_valid": defaultdict(int),
            "total": defaultdict(int),
        }
        self.report_available = False
        if isinstance(spec, str):
            spec = Path(spec)
        if isinstance(spec, Path):
            self.specfile = spec
            fmt = spec.suffix[1:]
            if fmt not in SUPPORTED_FORMATS:
                raise ValueError(f"adtl specification format not supported: {fmt}")
            with spec.open("rb") as fp:
                self.spec = SUPPORTED_FORMATS[fmt](fp)
        else:
            self.spec = spec
        self.header = self.spec.get("adtl", {})
        if self.specfile:
            self.include_defs = [
                relative_path(self.specfile, definition_file)
                for definition_file in self.header.get("include-def", [])
            ] + self.include_defs
        self.defs = self.header.get("defs", {})
        if self.include_defs:
            for definition_file in self.include_defs:
                self.defs.update(read_definition(definition_file))
        self.spec = expand_refs(self.spec, self.defs)

        self.validate_spec()
        for table in (t for t in self.tables if self.tables[t]["kind"] == "oneToMany"):
            self.spec[table] = expand_for(self.spec[table])
        for table in self.tables:
            if self.tables[table].get("groupBy"):
                self.data[table] = defaultdict(dict)
            else:
                self.data[table] = []
            if schema := self.tables[table].get("schema"):
                optional_fields = self.tables[table].get("optional-fields")
                if schema.startswith("http"):
                    try:
                        res = requests.get(schema)
                        if res.status_code != 200:
                            logging.warning(
                                f"Could not fetch schema for table {table!r}, will not "
                                "validate"
                            )
                            continue
                    except ConnectionError:  # pragma: no cover
                        logging.warning(
                            f"Could not fetch schema for table {table!r}, will not "
                            "validate"
                        )
                        continue
                    self.schemas[table] = make_fields_optional(
                        res.json(), optional_fields
                    )
                else:  # local file
                    with (self.specfile.parent / schema).open() as fp:
                        self.schemas[table] = make_fields_optional(
                            json.load(fp), optional_fields
                        )
                self.date_fields.extend(get_date_fields(self.schemas[table]))
                self.validators[table] = fastjsonschema.compile(self.schemas[table])

        self._set_field_names()

    @lru_cache
    def ctx(self, attribute: str):
        return {
            "is_date": attribute in self.date_fields,
            "defaultDateFormat": self.header.get(
                "defaultDateFormat", DEFAULT_DATE_FORMAT
            ),
            "skip_pattern": (
                re.compile(self.header.get("skipFieldPattern"))
                if self.header.get("skipFieldPattern")
                else False
            ),
            "returnUnmatched": self.header.get("returnUnmatched", False),
        }

    def validate_spec(self):
        "Raises exceptions if specification is invalid"
        for required in ["tables", "name", "description"]:
            if required not in self.header:
                raise ValueError(f"Specification header requires key: {required}")
        self.tables = self.header["tables"]
        self.name = self.header["name"]
        self.description = self.header["description"]

        for table in self.tables:
            aggregation = self.tables[table].get("aggregation")
            group_field = self.tables[table].get("groupBy")
            kind = self.tables[table].get("kind")
            if kind is None:
                raise ValueError(
                    f"Required 'kind' attribute within 'tables' not present for {table}"
                )
            if group_field is not None and aggregation != "lastNotNull":
                raise ValueError(
                    "groupBy needs aggregation=lastNotNull to be set for table: "
                    f"{table}"
                )

    def _set_field_names(self):
        for table in self.tables:
            if table not in self.spec:
                raise ValueError(
                    f"Parser specification missing required '{table}' element"
                )
            if self.tables[table].get("kind") != "oneToMany":
                self.fieldnames[table] = sorted(list(self.spec[table].keys()))
            else:
                if table not in self.schemas:
                    print(
                        f"Warning: no schema found for {table!r}, field names may be "
                        "incomplete!"
                    )
                    self.fieldnames[table] = list(
                        self.tables[table].get("common", {}).keys()
                    ) + sorted(
                        list(set(sum([list(m.keys()) for m in self.spec[table]], [])))
                    )
                else:
                    self.fieldnames[table] = sorted(self.schemas[table]["properties"])
                if commonMappings := self.tables[table].get("common", {}):
                    for match in self.spec[table]:
                        match.update(commonMappings)

    def _default_if(self, table: str, rule: StrDict):
        """
        Default behaviour for oneToMany table, row not displayed if there's an empty
        string or values not mapped in the rule.
        """

        data_options = [
            option["required"][0] for option in self.schemas[table]["oneOf"]
        ]

        option = set(data_options).intersection(rule.keys()).pop()

        if "combinedType" not in rule[option]:
            field = rule[option]["field"]
            if "values" in rule[option]:
                values = rule[option]["values"]
                if "can_skip" in rule[option]:
                    if_rule = {"any": [{field: v, "can_skip": True} for v in values]}
                else:
                    if_rule = {"any": [{field: v} for v in values]}
            elif "can_skip" in rule[option]:
                if_rule = {field: {"!=": ""}, "can_skip": True}
            else:
                if_rule = {field: {"!=": ""}}
        else:
            assert rule[option]["combinedType"] in [
                "any",
                "all",
                "firstNonNull",
                "set",
                "list",
                "min",
                "max",
            ], f"Invalid combinedType: {rule[option]['combinedType']}"
            rules = rule[option]["fields"]

            def create_if_rule(rule):
                field = rule["field"]
                values = rule.get("values", [])
                can_skip = rule.get("can_skip", False)

                if_condition = {}

                if values and can_skip:
                    if_condition = [{field: v, "can_skip": True} for v in values]
                elif values:
                    if_condition = [{field: v} for v in values]
                elif can_skip:
                    if_condition[field] = {"!=": ""}
                    if_condition["can_skip"] = True
                    if_condition = [if_condition]
                else:
                    if_condition[field] = {"!=": ""}
                    if_condition = [if_condition]

                return if_condition

            if_rule = {"any": sum(map(create_if_rule, rules), [])}

        rule["if"] = if_rule
        return rule

    def update_table(self, table: str, row: StrDict):
        """Updates table with a new row

        Args:
            table: Table to update
            row: Dictionary with keys as field names and values as field values
        """

        group_field = self.tables[table].get("groupBy")
        kind = self.tables[table].get("kind")
        if group_field:
            group_key = get_value(row, self.spec[table][group_field])
            for attr in self.spec[table]:
                value = get_value(row, self.spec[table][attr], self.ctx(attr))
                # Check against all null elements, for combinedType=set/list, null is []
                if value is not None and value != []:
                    if attr not in self.data[table][group_key].keys():
                        # if data for this field hasn't already been captured
                        self.data[table][group_key][attr] = value

                    else:
                        if "combinedType" in self.spec[table][attr]:
                            combined_type = self.spec[table][attr]["combinedType"]
                            existing_value = self.data[table][group_key][attr]

                            if combined_type in ["all", "any", "min", "max"]:
                                values = [existing_value, value]
                                # normally calling eval() is a bad idea, but here
                                # values are restricted, so okay
                                self.data[table][group_key][attr] = eval(combined_type)(
                                    values
                                )
                            elif combined_type in ["list", "set"]:
                                if combined_type == "set":
                                    self.data[table][group_key][attr] = list(
                                        set(existing_value + value)
                                    )
                                else:
                                    self.data[table][group_key][attr] = (
                                        existing_value + value
                                    )
                            elif combined_type == "firstNonNull":
                                # only use the first value found
                                pass
                        else:
                            # otherwise overwrite?
                            logging.debug(
                                f"Multiple rows of data found for {attr} without a"
                                " combinedType listed. Data being overwritten."
                            )
                            self.data[table][group_key][attr] = value

        elif kind == "oneToMany":
            for match in self.spec[table]:
                if "if" not in match:
                    match = self._default_if(table, match)
                if parse_if(row, match["if"], self.ctx):
                    self.data[table].append(
                        remove_null_keys(
                            {
                                attr: get_value(row, match[attr], self.ctx(attr))
                                for attr in set(match.keys()) - {"if"}
                            }
                        )
                    )
        elif kind == "constant":  # only one row
            self.data[table] = [self.spec[table]]
        else:
            self.data[table].append(
                remove_null_keys(
                    {
                        attr: get_value(row, self.spec[table][attr], self.ctx(attr))
                        for attr in self.spec[table]
                    }
                )
            )

    def parse(self, file: str, encoding: str = "utf-8", skip_validation=False):
        """Transform file according to specification

        Args:
            file: Source file to transform
            encoding: Source file encoding
            skip_validation: Whether to skip validation, default off

        Returns:
            adtl.Parser: Returns an instance of itself, updated with the parsed tables
        """
        with open(file, encoding=encoding) as fp:
            reader = csv.DictReader(fp)
            return self.parse_rows(
                (
                    tqdm(
                        reader,
                        desc=f"[{self.name}] parsing {Path(file).name}",
                    )
                    if not self.quiet
                    else reader
                ),
                skip_validation=skip_validation,
            )

    def parse_rows(self, rows: Iterable[StrDict], skip_validation=False):
        """Transform rows from an iterable according to specification

        Args:
            rows: Iterable of rows, specified as a dictionary of
                    (field name, field value) pairs
            skip_validation: Whether to skip validation, default off

        Returns:
            adtl.Parser: Returns an instance of itself, updated with the parsed tables
        """
        for row in rows:
            for table in self.tables:
                try:
                    self.update_table(table, row)
                except ValueError:  # pragma: no cover
                    print(
                        "\n".join(
                            [
                                f"{key} = {value}"
                                for key, value in row.items()
                                if value not in ["", None]
                            ]
                        )
                    )
                    raise
        self.report_available = not skip_validation
        if not skip_validation:
            for table in self.validators:
                for row in self.read_table(table):
                    self.report["total"][table] += 1
                    try:
                        self.validators[table](row)
                        row["adtl_valid"] = True
                        self.report["total_valid"][table] += 1
                    except fastjsonschema.exceptions.JsonSchemaValueException as e:
                        row["adtl_valid"] = False
                        row["adtl_error"] = e.message
                        self.report["validation_errors"][table].update([e.message])
        return self

    def clear(self):
        "Clears parser state"
        self.data = {}

    def read_table(self, table: str) -> Iterable[StrDict]:
        """Returns parsed table

        Args:
            table: Table to read

        Returns:
            Iterable of transformed rows in table
        """
        if table not in self.tables:
            raise ValueError(f"Invalid table: {table}")
        if "groupBy" in self.tables[table]:
            for i in self.data[table]:
                yield self.data[table][i]
        else:
            for row in self.data[table]:
                yield row

    def write_csv(self, table: str, output: str | None = None) -> str | None:
        """Writes to output as CSV a particular table

        Args:
            table: Table that should be written to CSV
            output: (optional) Output file name. If not specified, defaults to parser
                    name + table name with a csv suffix.
        """

        def writerows(fp, table):
            writer = csv.DictWriter(
                fp,
                fieldnames=(
                    ["adtl_valid", "adtl_error"] if table in self.validators else []
                )
                + self.fieldnames[table],
            )
            writer.writeheader()
            for row in self.read_table(table):
                writer.writerow(row)
            return fp

        if output:
            with open(output, "w") as fp:
                writerows(fp, table)
            return None
        else:
            buf = io.StringIO()
            return writerows(buf, table).getvalue()

    def write_parquet(self, table: str, output: str | None = None) -> str | None:
        """Writes to output as parquet a particular table

        Args:
            table: Table that should be written to parquet
            output: (optional) Output file name. If not specified, defaults to parser
                    name + table name with a parquet suffix.
        """

        try:
            import polars as pl
        except ImportError:
            raise ImportError(
                "Parquet output requires the polars library. "
                "Install with 'pip install polars'"
            )

        # Read the table data
        data = list(self.read_table(table))

        # Convert data to Polars DataFrame
        df = pl.DataFrame(data, infer_schema_length=len(data))

        if table in self.validators:
            valid_cols = [c for c in ["adtl_valid", "adtl_error"] if c in df.columns]
            df_validated = df.select(
                valid_cols
                + [
                    *[
                        col
                        for col in df.columns
                        if (col != "adtl_valid" and col != "adtl_error")
                    ],  # All other columns, in their original order
                ]
            )
        else:
            df_validated = df

        if output:
            df_validated.write_parquet(output)
        else:
            buf = io.BytesIO()
            df_validated.write_parquet(buf)
            return buf.getvalue()

    def show_report(self):
        "Shows report with validation errors"
        if self.report_available:
            print("\n|table       \t|valid\t|total\t|percentage_valid|")
            print("|---------------|-------|-------|----------------|")
            for table in self.report["total"]:
                print(
                    f"|{table:14s}\t|{self.report['total_valid'][table]}\t"
                    f"|{self.report['total'][table]}\t"
                    f"|{self.report['total_valid'][table] / self.report['total'][table]:%} |"  # noqa:E501
                )
            print()
            for table in self.report["validation_errors"]:
                print(f"## {table}\n")
                for message, count in self.report["validation_errors"][
                    table
                ].most_common():
                    print(f"* {count}: {message}")
                print()

    def save(
        self, output: str | None = None, format: Literal["csv", "parquet"] = "csv"
    ):
        """Saves all tables to CSV

        Args:
            output: (optional) Filename prefix that is used for all tables
        """

        if format == "parquet":
            for table in self.tables:
                self.write_parquet(table, f"{output}-{table}.parquet")

        elif format == "csv":
            for table in self.tables:
                self.write_csv(table, f"{output}-{table}.csv")

        else:
            raise ValueError(f"'Parser.save()': Invalid format: {format}")



"Contains all functions that call Google's Gemini API."

from __future__ import annotations

import json

import google.generativeai as gemini

from .base_llm import LLMBase
from .data_structures import ColumnDescriptionRequest, MappingRequest, ValuesRequest


class GeminiLanguageModel(LLMBase):
    def __init__(self, api_key, model: str = "gemini-1.5-flash"):
        gemini.configure(api_key=api_key)
        self.client = gemini.GenerativeModel(model)
        self.model = model

    def get_definitions(self, headers: list[str], language: str) -> dict[str, str]:
        """
        Get the definitions of the columns in the dataset using the Gemini API.
        """
        result = self.client.generate_content(
            [
                (
                    "You are an expert at structured data extraction. "
                    "The following is a list of headers from a data file in "
                    f"{language}, some containing shortened words or abbreviations. "
                    "Translate them to english. "
                    "Return a list of (original header, translation) pairs, using the given structure."  # noqa
                    "Preserve special characters such as accented letters and hyphens."
                ),
                f"{headers}",
            ],
            generation_config=gemini.GenerationConfig(
                response_mime_type="application/json",
                response_schema=ColumnDescriptionRequest,
            ),
        )
        descriptions = ColumnDescriptionRequest.model_validate(
            json.loads(result.text)
        ).field_descriptions
        return descriptions

    def map_fields(
        self, source_fields: list[str], target_fields: list[str]
    ) -> MappingRequest:
        """
        Calls the Gemini API to generate a draft mapping between two datasets.
        """
        result = self.client.generate_content(
            [
                (
                    "You are an expert at structured data extraction. "
                    "You will be given two lists of phrases, one is the headers for a "
                    "target data file, and the other a set of descriptions for columns "
                    "of source data. "
                    "Match each target header to the best matching source description, "
                    "but match a header to None if a good match does not exist. "
                    "Preserve special characters such as accented letters and hyphens."
                    "Return the matched target headers and source descriptions using the provided structure."  # noqa
                ),
                (
                    f"These are the target headers: {target_fields}\n"
                    f"These are the source descriptions: {source_fields}"
                ),
            ],
            generation_config=gemini.GenerationConfig(
                response_mime_type="application/json",
                response_schema=MappingRequest,
            ),
        )
        return MappingRequest.model_validate(json.loads(result.text))

    def map_values(
        self, values: list[tuple[str, set[str], list[str | None] | None]], language: str
    ) -> ValuesRequest:
        """
        Calls the Gemini API to generate a set of value mappings for the fields.
        """
        result = self.client.generate_content(
            [
                (
                    "You are an expert at structured data extraction. "
                    "You will be given a list of tuples, where each tuple contains "
                    "three sets of string values. "
                    "The first set contains field names for a dataset."
                    "The second set contains values from a source dataset in "
                    f"{language}, and the third set contains target values for an "
                    "english-language transformed dataset. "
                    "Match all the values in the second set to the appropriate values "
                    "in the third set. "
                    "Return a list of dictionaries, where each dictionary contains the "
                    "field name as a key, and a dictionary containing "
                    "source values as keys, and the target text as values, "
                    "as the values. For example, the result should look like this: "
                    "[{'field_name_1': {'source_value_a': 'target_value_a', "
                    "'source_value_b': 'target_value_b'}, 'field_name_2':{...}]"
                    "using the provided structure."
                    "Preserve special characters such as accented letters and hyphens."
                ),
                f"These are the field, source, target value sets: {values}",
            ],
            generation_config=gemini.GenerationConfig(
                response_mime_type="application/json",
                response_schema=ValuesRequest,
            ),
        )
        return ValuesRequest.model_validate(json.loads(result.text))



"""
Generate TOML parser from intermediate CSV file
"""

from __future__ import annotations

import argparse
import json
import logging
from pathlib import Path
from typing import Any

import pandas as pd

from .toml_writer import dump
from .util import DEFAULT_CONFIG, parse_choices, read_config_schema, read_data


def adtl_header(
    name: str,
    description: str,
    tables_schemas: dict,
    definitions: dict = {},
):
    "The ADTL-specific header for the TOML file"
    schemas = {}
    for table in tables_schemas:
        schemas[table] = {
            "kind": "oneToOne",
            "schema": f"{tables_schemas[table]}",
        }

    return {
        "adtl": {
            "name": name,
            "description": description,
            "returnUnmatched": True,
            "tables": schemas,
            **{"defs": definitions},
        }
    }


def refs_defs(choices, num_refs):
    references = {}
    definitions = {}

    top_mappings = choices[choices > 1][:num_refs].index

    # only add one boolean map for simplicity
    boolean_map_found = False
    for mapping in top_mappings:
        if boolean_map_found and True in mapping.values():
            continue
        if True in mapping.values():
            references[json.dumps(mapping, sort_keys=True)] = "Y/N/NK"
            definitions["Y/N/NK"] = {
                "caseInsensitive": True,
                "values": mapping,
            }
            boolean_map_found = True
            continue
        c = mapping
        name = "/".join(map(str, c.values()))
        references[json.dumps(mapping, sort_keys=True)] = name
        definitions[name] = {"values": c, "caseInsensitive": True}

    return references, definitions


class ParserGenerator:
    """
    Class for creating a TOML parser from an intermediate CSV file.

    Use `create_parser()` to write out the TOML parser file, as the function equivalent
    of the command line `create-parser` script.

    Parameters
    ----------
    mappings : pd.DataFrame | str | Path
        The intermediate CSV file created by `create_mapping.py`
    schema_path : Path | str
        The path to the folder containing all the schema files
    parser_name : str
        The name of the parser
    description : str, optional
        The description of the parser
    config : Path, optional
        The path to the configuration file to use if not using the default configuration
    transformation_tool : "ADTL"
        The transformation tool which will use the generated parser to convert data
        (only ADTL is currently supported)
    """

    def __init__(
        self,
        mappings: pd.DataFrame | str | Path,
        schema_path: Path | str,
        parser_name: str,
        description: str | None = None,
        config: Path | None = None,
        transformation_tool: str = "ADTL",
    ):
        self.mappings = read_data(mappings, "A mapping file")

        self.schema_path = Path(schema_path)
        self.parser_name = parser_name
        self.parser_description = description or parser_name

        self.config = read_config_schema(
            config or Path(Path(__file__).parent, DEFAULT_CONFIG)
        )
        self.tables = self.config["schemas"].keys()

        self.schemas = {
            t: read_config_schema(Path(schema_path, self.config["schemas"][t]))
            for t in self.tables
        }

        if transformation_tool != "ADTL":
            raise NotImplementedError(
                "Only ADTL is supported as a transformation tool, "
                f"not {transformation_tool}"
            )
        else:
            self.header = adtl_header(
                self.parser_name,
                self.parser_description,
                {
                    t: self.schema_path / Path(self.config["schemas"][t])
                    for t in self.tables
                },
                self.references_definitions[1],
            )

    @property
    def parsed_choices(self) -> pd.Series:
        """Returns the mapped values for each target field"""
        try:
            return self._parsed_choices
        except AttributeError:
            self._parsed_choices = self.mappings.value_mapping.map(parse_choices)
            self._parsed_choices.index = self.mappings.target_field
            return self._parsed_choices

    @property
    def references_definitions(self) -> tuple[dict[str, str], dict[str, dict]]:
        """Finds and returns the references and definitions for the mappings"""
        try:
            return self._references_definitions
        except AttributeError:
            # use value_counts() on parsed_choices normalise various flavours of Y/N/NK
            value_counts = self.parsed_choices.value_counts()

            self._references_definitions = refs_defs(
                value_counts, self.config["num_refs"]
            )
            return self._references_definitions

    def schema_fields(self, table: str):
        """Returns all the fields for `table` and their properties"""
        return self.schemas[table]["properties"]

    def field_types(self, table) -> dict[str, list[str]]:
        """Returns the field types of the target schema"""
        try:
            return self._field_types
        except AttributeError:
            s = self.schema_fields(table)
            self._field_types = {
                f: s[f].get("type", ["string", "null"]) for f in s.keys()
            }
            return self._field_types

    def single_field_mapping(self, table, match: pd.DataFrame) -> dict[str, Any]:
        """Make a single field mapping from a single row of the mappings dataframe"""

        choices = self.parsed_choices[match.target_field]

        out = {"field": match.source_field, "description": match.source_description}
        references = self.references_definitions[0]
        if choices:
            if (choice_key := json.dumps(choices, sort_keys=True)) in references:
                out["ref"] = references[choice_key]
            else:
                out["values"] = choices
                out["caseInsensitive"] = True

            if "array" in self.field_types(table)[match.target_field]:
                out["type"] = "enum_list"

        return out

    def make_toml_table(self, table: str) -> dict[str, Any]:
        """Make single TOML table from mappings"""

        outmap = {}

        for field, field_matches in self.mappings.groupby("target_field"):
            if len(field_matches) == 1:  # single field
                if not any(field_matches["source_field"].isna()):
                    outmap[field] = self.single_field_mapping(
                        table, field_matches.iloc[0]
                    )

            else:  # combinedType
                raise NotImplementedError("CombinedType not supported")

        # check for missing required fields
        schema = self.schemas[table]
        if "required" in schema:
            for field in schema["required"]:
                if field not in outmap:
                    logging.warning(
                        f"Missing required field {field} in {table} schema."
                        " Adding empty field..."
                    )
                    outmap[field] = ""

        return {table: outmap}

    def make_toml(self) -> dict[str, Any]:
        """
        Takes the csv mapping file from `create_mapping` and writes out a TOML parser

        Generates a TOML parser for use with ADTL using the intermediate CSV file from
        by `create_mapping`. This will generate a TOML file that can be used to parse
        raw data into the format expected by the schema.

        Returns
        -------
        dict
            Dictionary containing the TOML parser data, ready to be written out.
        """

        data = self.header
        for table in self.tables:
            data.update(self.make_toml_table(table))
        return data

    def write_toml(self, data: dict[str, Any], output: str = None):
        """
        Write a dictionary structure to a TOML file, using `output` as the filename if
        provided.

        Parameters
        ----------
        data : dict
            Dictionary containing the TOML parser data
        output : str, optional
            Filename to write the TOML data to. Defaults to the name of the parser.
        """
        if not output:
            output = f"{self.parser_name}.toml"
        with open(output, "wb") as fp:
            dump(data, fp)

    def create_parser(self, file_name: str = None):
        """
        Main function to create the TOML parser from the intermediate CSV file.
        """
        toml_data = self.make_toml()
        self.write_toml(toml_data, output=file_name)


def create_parser(
    mappings: pd.DataFrame | str,
    schema_path: Path,
    parser_name: str,
    description: str | None = None,
    config=DEFAULT_CONFIG,
):
    """
    Takes the csv mapping file created by `create_mapping` and writes out a TOML parser

    Generates a TOML parser for use with ADTL from the intermediate CSV file generated
    by `create_mapping`. This will generate a TOML file that can be used to parse raw
    data into the format expected by the schema.

    Parameters
    ----------
    mappings
        Path to the CSV file containing the mappings
    schema_path
        Path to the schema file
    parser_name
        Name of the parser to create
    description
        Description of the parser. Defaults to the parser name.
    config
        Path to the configuration file to use. Default is `config/autoparser.toml`.

    Returns
    -------
    None
    """
    ParserGenerator(
        mappings,
        schema_path,
        parser_name,
        description,
        Path(config),
    ).create_parser(parser_name)


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Make TOML from intermediate CSV file created by create_mapping.py",
        prog="autoparser create-parser",
    )
    parser.add_argument("mappings", help="Mapping file to create TOML from", type=str)
    parser.add_argument("schema_path", help="Path where schemas are located")
    parser.add_argument(
        "-o",
        "--output",
        help="Name of the parser to output (default=globalhealth)",
        default="globalhealth",
    )
    parser.add_argument("--description", help="Description of the parser")
    parser.add_argument(
        "-c",
        "--config",
        help=f"Configuration file to use (default={DEFAULT_CONFIG})",
        type=Path,
    )
    args = parser.parse_args(argv)

    schema_path = Path(args.schema_path)

    ParserGenerator(
        args.mappings,
        schema_path,
        args.output,
        args.description or None,
        args.config or None,
    ).create_parser()


if __name__ == "__main__":
    main()



"""Stores the data structures for using with LLM API's"""

from __future__ import annotations

from pydantic import BaseModel

# target classes for generating descriptions


class SingleField(BaseModel):
    field_name: str
    translation: str


class ColumnDescriptionRequest(BaseModel):
    field_descriptions: list[SingleField]


# target classes for matching fields
class SingleMapping(BaseModel):
    target_field: str
    source_description: str | None


class MappingRequest(BaseModel):
    targets_descriptions: list[SingleMapping]


# target classes for matching values to enum/boolean options
class ValueMapping(BaseModel):
    source_value: str
    target_value: str | None


class FieldMapping(BaseModel):
    field_name: str
    mapped_values: list[ValueMapping]


class ValuesRequest(BaseModel):
    values: list[FieldMapping]



import argparse
import importlib.metadata
import json

from adtl.parser import Parser
from adtl.python_interface import parse

__all__ = ["Parser", "parse"]

__version__ = importlib.metadata.version("adtl")


def main(argv=None):
    cmd = argparse.ArgumentParser(
        prog="adtl",
        description="Transforms and validates data into CSV given a specification",
    )
    cmd.add_argument(
        "spec",
        help="specification file to use",
    )
    cmd.add_argument("file", help="file to read in")
    cmd.add_argument(
        "-o", "--output", help="output file, if blank, writes to standard output"
    )
    cmd.add_argument(
        "--encoding", help="encoding input file is in", default="utf-8-sig"
    )
    cmd.add_argument(
        "--parquet", help="output file is in parquet format", action="store_true"
    )
    cmd.add_argument(
        "-q",
        "--quiet",
        help="quiet mode - decrease verbosity, disable progress bar",
        action="store_true",
    )
    cmd.add_argument("--save-report", help="save report in JSON format")
    cmd.add_argument(
        "--include-def",
        action="append",
        help="include external definition (TOML or JSON)",
    )
    cmd.add_argument("--version", action="version", version="%(prog)s " + __version__)
    args = cmd.parse_args(argv)
    include_defs = args.include_def or []
    spec = Parser(args.spec, include_defs=include_defs, quiet=args.quiet)

    # check for incompatible options
    if spec.header.get("returnUnmatched") and args.parquet:
        raise ValueError("returnUnmatched and parquet options are incompatible")

    # run adtl
    adtl_output = spec.parse(args.file, encoding=args.encoding)
    adtl_output.save(args.output or spec.name, "parquet" if args.parquet else "csv")
    if args.save_report:
        adtl_output.report.update(
            dict(
                encoding=args.encoding,
                include_defs=include_defs,
                file=args.file,
                parser=args.spec,
            )
        )
        with open(args.save_report, "w") as fp:
            json.dump(adtl_output.report, fp, sort_keys=True, indent=2)
    else:
        adtl_output.show_report()


if __name__ == "__main__":
    main()



from __future__ import annotations

import hashlib
import logging
import re
import warnings
from datetime import datetime
from typing import Any, Callable, Iterable, Union

import pint

import adtl.transformations as tf
from adtl.transformations import AdtlTransformationWarning

StrDict = dict[str, Any]
Rule = Union[str, StrDict]
Context = Union[dict[str, Union[bool, int, str, list[str]]], None]

# helper functions


def matching_fields(fields: list[str], pattern: str) -> list[str]:
    "Returns fields matching pattern"
    compiled_pattern = re.compile(pattern)
    return [f for f in fields if compiled_pattern.match(f)]


def flatten(xs):
    """
    Flatten a list of lists +-/ non-list items
    e.g.
    [None, ['Dexamethasone']] -> [None, 'Dexamethasome']
    """
    for x in xs:
        if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
            yield from flatten(x)
        else:
            yield x


def hash_sensitive(value: str) -> str:
    """Hashes sensitive values. This is not generally sufficient for
    anonymisation, as the value still serves as a unique identifier,
    but is better than storing the value unprocessed."""
    return hashlib.sha256(str(value).encode("utf-8")).hexdigest()


def skip_field(row: StrDict, rule: StrDict, ctx: Context = None):
    "Returns True if the field is missing and allowed to be skipped"
    if rule.get("can_skip"):
        return rule["field"] not in row
    if ctx and ctx.get("skip_pattern") and ctx.get("skip_pattern").match(rule["field"]):
        return rule["field"] not in row
    return False


def apply_fuction(value, row: StrDict, rule: StrDict, ctx: Context):
    # apply data transformations.
    transformation = rule["apply"]["function"]
    params = None
    if "params" in rule["apply"]:
        params = []
        for i in range(len(rule["apply"]["params"])):
            if isinstance(rule["apply"]["params"][i], str) and rule["apply"]["params"][
                i
            ].startswith("$"):
                params.append(row[rule["apply"]["params"][i][1:]])
            elif isinstance(rule["apply"]["params"][i], list):
                param = [
                    (
                        row[rule["apply"]["params"][i][j][1:]]
                        if (
                            isinstance(rule["apply"]["params"][i][j], str)
                            and rule["apply"]["params"][i][j].startswith("$")
                        )
                        else rule["apply"]["params"][i][j]
                    )
                    for j in range(len(rule["apply"]["params"][i]))
                ]
                params.append(param)
            else:
                params.append(rule["apply"]["params"][i])

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("error", category=AdtlTransformationWarning)
            if params:
                value = getattr(tf, transformation)(value, *params)
            else:
                value = getattr(tf, transformation)(value)
    except AttributeError:
        raise AttributeError(
            f"Error using a data transformation: Function {transformation} "
            "has not been defined."
        )
    except AdtlTransformationWarning as e:
        if ctx and ctx.get("returnUnmatched"):
            warnings.warn(str(e), AdtlTransformationWarning)
            return value
        else:
            logging.error(str(e))
            return None
    return value


def convert_values(value, rule: StrDict, ctx: Context) -> str | list[str | None] | None:
    if rule.get("type") == "enum_list":
        try:
            value = [v.lstrip(" ").rstrip(" ") for v in value.strip("[]").split(",")]
            new_rule = {k: v for k, v in rule.items() if k != "type"}
            value = [convert_values(v, new_rule, ctx) for v in value]
            return value
        except Exception as e:
            logging.debug(f"Error converting {value} to a list: {e}")
            return value

    if rule.get("caseInsensitive") and isinstance(value, str):
        value = value.lower().lstrip(" ").rstrip(" ")
        rule["values"] = {k.lower(): v for k, v in rule["values"].items()}

    if rule.get("ignoreMissingKey") or (ctx and ctx.get("returnUnmatched")):
        value = rule["values"].get(value, value)
    else:
        value = rule["values"].get(value)

    # recheck if value is empty after mapping (use to map values to None)
    return None if value == "" else value


# main functions


def get_value(row: StrDict, rule: Rule, ctx: Context = None) -> Any:
    """Gets value from row using rule

    Same as get_value_unhashed(), except it hashes if sensitive = True in rule.
    This function should be used instead of get_value_unhashed() for
    application code.
    """
    value = get_value_unhashed(row, rule, ctx)
    if isinstance(rule, dict) and rule.get("sensitive") and value is not None:
        return hash_sensitive(value)
    if not isinstance(value, str):
        return value
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            return value


def get_value_unhashed(row: StrDict, rule: Rule, ctx: Context = None) -> Any:
    """Gets value from row using rule (unhashed)

    Unlike get_value() this function does NOT hash sensitive data
    and should not be called directly, except for debugging. Use
    get_value() instead.
    """
    if not isinstance(rule, dict) or isinstance(rule, list):
        # not a container, is constant
        return rule
    # Check whether field is present if it's allowed to be passed over
    if "field" in rule:
        # do not check for condition if field is missing
        if skip_field(row, rule, ctx):
            return None
        # do not parse field if condition is not met
        if "if" in rule and not parse_if(row, rule["if"]):
            return None
        value = row[rule["field"]]
        if "apply" in rule:
            value = apply_fuction(value, row, rule, ctx)
        if value == "":
            return None
        if "values" in rule:
            value = convert_values(value, rule, ctx)
        # Either source_unit / unit OR source_date / date triggers conversion
        # do not parse units if value is empty
        if "source_unit" in rule and "unit" in rule:
            assert "source_date" not in rule and "date" not in rule
            source_unit = get_value(row, rule["source_unit"])
            unit = rule["unit"]
            if not isinstance(source_unit, str):
                logging.debug(
                    f"Error converting source_unit {source_unit} to {unit!r} with "
                    "rule: {rule}, defaulting to assume source_unit is {unit}"
                )
                return float(value)
            try:
                value = pint.Quantity(float(value), source_unit).to(unit).m
            except ValueError:
                if ctx and ctx.get("returnUnmatched"):
                    logging.debug(f"Could not convert {value} to a floating point")
                    return value
                raise ValueError(f"Could not convert {value} to a floating point")
        if "source_date" in rule or (ctx and ctx.get("is_date")):
            assert "source_unit" not in rule and "unit" not in rule
            target_date = rule.get("date", "%Y-%m-%d")
            source_date = (
                get_value(row, rule["source_date"])
                if "source_date" in rule
                else ctx["defaultDateFormat"]
            )
            if source_date != target_date:
                try:
                    value = datetime.strptime(value, source_date).strftime(target_date)
                except (TypeError, ValueError):
                    logging.info(f"Could not parse date: {value}")
                    if ctx and ctx.get("returnUnmatched"):
                        return value
                    return None
        return value
    elif "combinedType" in rule:
        return get_combined_type(row, rule, ctx)
    else:
        raise ValueError(f"Could not return value for {rule}")


def parse_if(
    row: StrDict, rule: StrDict, ctx: Callable[[str], dict] = None, can_skip=False
) -> bool:
    "Parse conditional statements and return a boolean"

    n_keys = len(rule.keys())
    assert n_keys == 1 or n_keys == 2
    if n_keys == 2:
        assert "can_skip" in rule
        can_skip = True
    key = next(iter(rule.keys()))
    if key == "not" and isinstance(rule[key], dict):
        return not parse_if(row, rule[key], ctx, can_skip)
    elif key == "any" and isinstance(rule[key], list):
        return any(parse_if(row, r, ctx, can_skip) for r in rule[key])
    elif key == "all" and isinstance(rule[key], list):
        return all(parse_if(row, r, ctx, can_skip) for r in rule[key])
    try:
        attr_value = row[key]
    except KeyError:
        if can_skip is True:
            return False
        elif ctx:
            if skip_field(row, {"field": key}, ctx(key)):
                return False
        raise

    if isinstance(rule[key], dict):
        cmp = next(iter(rule[key]))
        value = rule[key][cmp]
        try:
            cast_value = type(value)(attr_value)
        except ValueError:
            logging.debug(
                f"Error when casting value {attr_value!r} with rule: {rule}, defaulting"
                " to False"
            )
            return False
        if cmp == ">":
            return cast_value > value
        elif cmp == ">=":
            return cast_value >= value
        elif cmp == "<":
            return cast_value < value
        elif cmp == "<=":
            return cast_value <= value
        elif cmp == "!=":
            return cast_value != value
        elif cmp in ["=", "=="]:
            return cast_value == value
        elif cmp == "=~":
            return bool(re.match(value, cast_value, re.IGNORECASE))
        else:
            raise ValueError(f"Unrecognized operand: {cmp}")
    elif isinstance(rule[key], set):  # common error, missed colon to make it a dict
        raise ValueError(
            f"if-subexpressions should be a dictionary, is a set: {rule[key]}"
        )
    else:
        value = rule[key]
        try:
            cast_value = type(value)(attr_value)
        except ValueError:
            logging.debug(
                f"Error when casting value {attr_value!r} with rule: {rule}, defaulting"
                " to False"
            )
            return False
        return cast_value == value


def get_combined_type(row: StrDict, rule: StrDict, ctx: Context = None):
    """Gets value from row for a combinedType rule

    A rule with the combinedType key combines multiple fields in the row
    to get the value. Thus this rule assumes that the combinedType fields
    do NOT have repeated (possibly different) values across the dataset.

    Example of dataset that will be handled correctly, with modliv and
    mildliver being the categorical indicators for moderate and mild
    liver disease respectively:

        subjid,modliv,mildliver,otherfield
        1,0,1,NA
        1,,,

    Example of dataset that will not be handled correctly:

        subjid,modliv,mildliver,otherfield
        1,0,,
        1,,1,

    For a combinedType rule to successfully run, all the field values should
    be present in the same row.
    """
    assert "combinedType" in rule
    combined_type = rule["combinedType"]
    rules = []
    # expand fieldPattern rules
    for r in rule["fields"]:
        if "fieldPattern" in r:
            for match in matching_fields(list(row.keys()), r.get("fieldPattern")):
                rules.append({"field": match, **r})
        else:
            rules.append(r)
    if combined_type in ["all", "any", "min", "max"]:
        values = [get_value(row, r, ctx) for r in rules]
        values = [v for v in values if v not in [None, ""]]
        # normally calling eval() is a bad idea, but here values are restricted, so okay
        return eval(combined_type)(values) if values else None
    elif combined_type == "firstNonNull":
        try:
            return next(
                filter(
                    lambda item: item is not None,
                    flatten([get_value(row, r, ctx) for r in rules]),
                )
            )
        except StopIteration:
            return None
    elif combined_type == "list" or combined_type == "set":
        excludeWhen = rule.get("excludeWhen")
        if excludeWhen not in [None, "false-like", "none"] and not isinstance(
            excludeWhen, list
        ):
            raise ValueError(
                "excludeWhen rule should be 'none', 'false-like', or a list of values"
            )

        values = flatten([get_value(row, r, ctx) for r in rules])
        if combined_type == "set":
            values = [*set(values)]
        if excludeWhen is None:
            return list(values)
        if excludeWhen == "none":
            return [v for v in values if v is not None]
        elif excludeWhen == "false-like":
            return [v for v in values if v]
        else:
            return [v for v in values if v not in excludeWhen]
    else:
        raise ValueError(f"Unknown {combined_type} in {rule}")



"""
Infer a data dictionary from a dataset.
"""

from __future__ import annotations

import argparse
import warnings
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd

from .util import (
    DEFAULT_CONFIG,
    check_matches,
    load_data_dict,
    read_config_schema,
    read_data,
    setup_llm,
)


class DictWriter:
    """
    Class for inferring a data dictionary based on a dataset. Will not store the data,
    only the created data dictionary.

    Use `create_dict()` to create a data dictionary, as the function equivalent
    of the command line `create-dict` script.

    `generate_descriptions()` will use an LLM to generate descriptions for the
    data dictionary, using only the column headers, NOT the data itself.

    Parameters
    ----------
    config
        The path to the configuration file to use if not using the default configuration
    """

    def __init__(
        self,
        config: Path | str | None = None,
        llm: str | None = None,
        api_key: str | None = None,
    ):
        if isinstance(config, str):
            config = Path(config)
        self.config = read_config_schema(
            config or Path(Path(__file__).parent, DEFAULT_CONFIG)
        )

        try:
            self.config["max_common_count"]
        except KeyError:
            raise ValueError("'max_common_count' not found in config file.")

        if llm and api_key:
            self.model = setup_llm(llm, api_key)
        else:
            self.model = None

    def _reset_dict_headers(
        self, config: dict[str:Any], data_dict: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Reset the headers of a data dictionary to those provided.

        Parameters
        ----------
        config
            Loaded configuration

        Returns
        -------
        pd.DataFrame
            Data dictionary with the headers reset
        """

        column_mappings = config["column_mappings"]
        data_dict.rename(columns=column_mappings, inplace=True)
        return data_dict

    def create_dict(self, data: pd.DataFrame | str) -> pd.DataFrame:
        """
        Create a basic data dictionary from a dataset.

        Creates a data dictionary from a dataset, including the field name, field type,
        and common values (defined as occuring more than 25 times in the columns).
        Also creates an empty column for field decriptions, which can either be added by
        hand later, or auto-generated with an LLM using `generate_descriptions()`.

        Parameters
        ----------
        data
            Path to a CSV or XLSX file, or a DataFrame, containing the raw data.

        Returns
        -------
        pd.DataFrame
            Data dictionary containing field names, field types, and common values.
        """

        df = read_data(data, "Data")

        names = df.columns
        types = [str(t) for t in df.dtypes]
        value_opts = {}

        # Get common value thresholds
        max_common_count = self.config["max_common_count"]
        min_common_freq = self.config.get("min_common_freq")

        # check the max count isn't > than 30% of the dataset
        calced_max_common_count = min(max_common_count, len(df) * 0.3)
        if calced_max_common_count < max_common_count:
            warnings.warn(
                f"Small Dataset: max_common_count of '{max_common_count}' is"
                f" too high for a dataset with {len(df)} rows.\n"
                f"Reducing to {calced_max_common_count} to avoid data "
                "identification issues.\n"
                "Setting the minimum frequency to 5% of the dataset."
            )
            max_common_count = calced_max_common_count
            min_common_freq = 0.05

        for j, i in enumerate(df.columns):
            values = df[i].value_counts(sort=True)

            # check for lists in the data
            if types[j] == "object" and any(
                "[" in x or "," in x for x in values.index.values if isinstance(x, str)
            ):
                # the values might be lists.
                list_col = df[i].apply(
                    lambda x: (
                        [v.lstrip(" ").rstrip(" ") for v in x.strip("[]").split(",")]
                        if isinstance(x, str)
                        else x
                    )
                )
                flat_col = [item for sublist in list_col.dropna() for item in sublist]
                values = pd.Series(flat_col).value_counts(sort=True)
                types[j] = "list"

            if min_common_freq:
                values = values[values > max(1, len(df) * min_common_freq)]
            value_opts[i] = np.nan
            if not values.empty and len(values) <= max_common_count:
                # drop any values with a frequency of 1
                values = values[values > 1]
                if not values.empty:
                    try:
                        value_opts[i] = f"{self.config['choice_delimiter']} ".join(
                            list(values.index.values)
                        )
                    except TypeError:
                        # This stops float values being given as 'common values'.
                        continue

        dd = pd.DataFrame(
            {
                "Field Name": names,
                "Description": [np.nan] * len(names),
                "Field Type": types,
                "Common Values": value_opts.values(),
            }
        )
        dd["Field Type"] = dd["Field Type"].map(
            {
                "object": None,
                "int64": "number",
                "float64": "number",
                "datetime64[ns]": "date",
                "boolean": "bool",
                "list": "list",
            }
        )
        dd["Field Type"] = dd["Field Type"].fillna("string")

        self.data_dictionary = dd

        return dd

    def generate_descriptions(
        self,
        language: str,
        data_dict: pd.DataFrame | str | None = None,
        key: str | None = None,
        llm: str | None = "openai",
    ) -> pd.DataFrame:
        """
        Generate descriptions for the columns in the dataset.

        Uses an LLM to auto-generate descriptions for a data dictionary based on the
        column headers.

        Parameters
        ----------
        language
            Language the column headers are in (e.g. french, spanish).
        data_dict
            Data dictionary containing the column headers, either as a dataframe or a
            path to the dictionary as a csv/xlsx file. Can be None if the data dict
            has already been created using `create_dict()`.
        key
            OpenAI API key.
        llm
            LLM API to call (currently only OpenAI is supported)

        Returns
        -------
        pd.DataFrame
            Data dictionary with descriptions added
        """
        if data_dict is None:
            try:
                data_dict = self.data_dictionary
            except AttributeError:
                raise ValueError(
                    "No data dictionary found. Please create a data dictionary first."
                )

        df = load_data_dict(self.config, data_dict)

        if not self.model:
            self.model = setup_llm(llm, key)

        headers = df.source_field

        descriptions = self.model.get_definitions(list(headers), language)

        descriptions = {d.field_name: d.translation for d in descriptions}
        df_descriptions = pd.DataFrame(
            descriptions.items(), columns=["source_field_gpt", "description"]
        )

        descrip = pd.concat([headers, df_descriptions], axis=1)

        # check ordering is correct even if the return field names aren't quite the same
        # e.g. numbering has been stripped
        assert all(
            descrip.apply(
                lambda x: check_matches(x["source_field_gpt"], [x.source_field]),
                axis=1,
            )
        ), "Field names from the LLM don't match the originals."

        descrip.drop(columns=["source_field_gpt"], inplace=True)

        new_dd = pd.merge(df, descrip, on="source_field")
        new_dd["source_description"] = new_dd["description"]
        new_dd.drop(columns=["description"], inplace=True)

        new_dd = self._reset_dict_headers(self.config, new_dd)

        return new_dd


def create_dict(data: pd.DataFrame | str, config: Path | None = None) -> pd.DataFrame:
    """
    Create a basic data dictionary from a dataset.

    Creates a data dictionary from a dataset, including the field name, field type,
    and common values (defined as occuring more than 25 times in the columns).
    Also creates an empty column for field decriptions, which can either be added by
    hand later, or auto-generated with an LLM using `generate_descriptions()`.

    Parameters
    ----------
    data
        Path to a CSV or XLSX file, or a DataFrame, containing the raw data.
    config
        Path to the configuration file to use if not using the default configuration

    Returns
    -------
    pd.DataFrame
        Data dictionary containing field names, field types, and common values.
    """

    dd = DictWriter(config).create_dict(data)
    return dd


def generate_descriptions(
    data_dict: pd.DataFrame | str,
    language: str,
    key: str | None = None,
    llm: str | None = "openai",
    config: Path | None = None,
) -> pd.DataFrame:
    """
    Generate descriptions for the columns in the dataset.

    Uses an LLM to auto-generate descriptions for a data dictionary based on the
    column headers.

    Parameters
    ----------
    data_dict
        Data dictionary containing the column headers, either as a dataframe or a path
        to the dictionary as a csv/xlsx file.
    language
        Language the column headers are in (e.g. french, spanish).
    key
        OpenAI API key.
    llm
        LLM API to call (currently only OpenAI is supported)
    config
        Path to the configuration file to use if not using the default configuration

    Returns
    -------
    pd.DataFrame
        Data dictionary with descriptions added
    """

    dd = DictWriter(config=config).generate_descriptions(language, data_dict, key, llm)

    return dd


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Generate a basic data dictionary from a dataset",
        prog="autoparser create-dict",
    )
    parser.add_argument("data", help="Data to create dictionary from")
    parser.add_argument(
        "language", help="Language of the raw data (e.g. 'fr', 'en', 'es')"
    )
    parser.add_argument(
        "-d",
        "--descriptions",
        help="Use an LLM to generate descriptions from file headers",
        action="store_true",
    )
    parser.add_argument(
        "-k", "--api-key", help="OpenAI API key to generate descriptions"
    )
    parser.add_argument("-l", "--llm", help="LLM API to use", default="openai")
    parser.add_argument(
        "-c",
        "--config",
        help=f"Configuration file to use (default={DEFAULT_CONFIG})",
        type=Path,
    )
    parser.add_argument(
        "-o", "--output", help="Name to use for output files", default="datadict"
    )

    args = parser.parse_args(argv)

    if args.descriptions and not args.api_key:
        raise ValueError("API key required for generating descriptions")

    df = create_dict(args.data, args.config)
    if args.descriptions:
        df = generate_descriptions(
            df, args.language, args.api_key, args.llm, args.config
        )

    df.to_csv(f"{args.output}.csv", index=False)


if __name__ == "__main__":
    main()



"""
Common utility functions for autoparser
"""

from __future__ import annotations

import difflib
import json
import re
from pathlib import Path
from typing import Any, Dict

import pandas as pd
import tomli

from adtl.autoparser.language_models.gemini import GeminiLanguageModel
from adtl.autoparser.language_models.openai import OpenAILanguageModel

DEFAULT_CONFIG = "config/autoparser.toml"


def read_config_schema(path: str | Path) -> Dict:
    if isinstance(path, str):
        path = Path(path)

    if path.suffix == ".json":
        return read_json(path)
    elif path.suffix == ".toml":
        with path.open("rb") as fp:
            return tomli.load(fp)
    else:
        raise ValueError(
            f"read_config_schema(): Unsupported file format: {path.suffix}"
        )


def read_json(file: str | Path) -> dict:
    if isinstance(file, str):
        file = Path(file)

    with file.open() as fp:
        return json.load(fp)


def read_data(file: str | Path | pd.DataFrame, file_type: str):
    """Reads in data/mapping files, which expect csv, excel or dataframe formats"""

    if isinstance(file, str):
        file = Path(file)
        if file.suffix == ".csv":
            return pd.read_csv(file)
        elif file.suffix == ".xlsx":
            return pd.read_excel(file)
        else:
            raise ValueError(f"Unsupported format (not CSV or XLSX): {file}")
    elif isinstance(file, pd.DataFrame):
        return file
    else:
        raise ValueError(
            f"{file_type} must be a path to a CSV or XLSX file, or a DataFrame"
        )


def parse_choices(s: str) -> Dict[str, Any] | None:
    """
    Takes the choices from llm as a string and turns into pairs.

    "oui=True, non=False, blah=None" -> {"oui": True, "non": False, "blah": ""}
    "vivant=alive, décédé=dead, " "=None" -> {"vivant": "alive", "décédé": "dead"}
    {2: True} -> None
    "" " = " ", poisson=fish" -> {"poisson": "fish"}
    ecouvillon+croûte=[swab, crust], ecouvillon=[swab]" ->
            {"ecouvillon+croûte": ["swab", "crust"], "ecouvillon": ["swab"]}

    """

    if not isinstance(s, str):
        return None

    split_str = re.split(r",(?!(?:[^\[]*\])|(?:[^\[]*\[[^\]]*$))", s)
    choices_list = [tuple(x.strip().split("=")) for x in split_str]
    if any(len(c) != 2 for c in choices_list):
        raise ValueError(
            f"autoparser: Invalid choices list for value mapping {choices_list!r}"
        )
    choices = dict(choices_list)

    for k, v in choices.copy().items():
        if v.lower() == "true":
            choices[k] = True
        if v.lower() == "false":
            choices[k] = False
        if v.lower() == "none":
            if k == "":
                choices.pop(k)
            else:
                choices[k] = ""
        if v == "" and k == "":
            choices.pop(k)
        if "[" and "]" in v:
            choices[k] = [i for i in v.strip("[]").replace(" ", "").split(",")]
    return choices


def load_data_dict(
    config: dict[str:Any], data_dict: str | Path | pd.DataFrame
) -> pd.DataFrame:
    if isinstance(data_dict, str):
        data_dict = Path(data_dict)
        if data_dict.suffix == ".csv":
            data_dict = pd.read_csv(data_dict)
        elif data_dict.suffix == ".xlsx":  # pragma: no cover
            data_dict = pd.read_excel(data_dict)
        else:
            raise ValueError(f"Unsupported format (not CSV or XLSX): {data_dict}")

    column_mappings = {v: k for k, v in config["column_mappings"].items()}
    data_dict.rename(columns=column_mappings, inplace=True)
    return data_dict


def setup_llm(provider, api_key):
    """
    Setup the LLM to use to generate descriptions.

    Separate from the __init__ method to allow for extra barrier between raw data &
    LLM.

    Parameters
    ----------
    key
        API key
    name
        Name of the LLM to use (currently only OpenAI and Gemini are supported)
    """
    if api_key is None:
        raise ValueError("API key required to set up an LLM")

    if provider == "openai":  # pragma: no cover
        return OpenAILanguageModel(api_key=api_key)
    elif provider == "gemini":  # pragma: no cover
        return GeminiLanguageModel(api_key=api_key)
    else:
        raise ValueError(f"Unsupported LLM provider: {provider}")


def check_matches(llm: str, source: list[str], cutoff=0.8) -> str | None:
    """
    Use to check if a string returned by an llm is a close enough match to the original
    source.

    Useful for checking or finding the original word if the LLM misspells it when
    returning results.

    Parameters
    ----------
    llm
        String returned by the LLM
    source
        List of strings to compare against (usually the original fields/descriptions
        from the previous step)
    """
    if not isinstance(source, list):
        raise ValueError(
            f"check matches: source must be a list of strings, got '{source}'"
        )
    matches = difflib.get_close_matches(llm, source, n=1, cutoff=cutoff)
    return matches[0] if matches else None



"Contains all functions that call OpenAI's API."

from __future__ import annotations

from openai import OpenAI

from .base_llm import LLMBase
from .data_structures import ColumnDescriptionRequest, MappingRequest, ValuesRequest


class OpenAILanguageModel(LLMBase):
    def __init__(self, api_key, model: str = "gpt-4o-mini"):
        self.client = OpenAI(api_key=api_key)
        self.model = model

    def get_definitions(self, headers: list[str], language: str) -> dict[str, str]:
        """
        Get the definitions of the columns in the dataset.
        """
        completion = self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": (
                        "You are an expert at structured data extraction. "
                        "The following is a list of headers from a data file in "
                        f"{language}, some containing shortened words or abbreviations. "  # noqa
                        "Translate them to english. "
                        "Return a list of (original header, translation) pairs, using the given structure."  # noqa
                    ),
                },
                {"role": "user", "content": f"{headers}"},
            ],
            response_format=ColumnDescriptionRequest,
        )
        descriptions = completion.choices[0].message.parsed.field_descriptions

        return descriptions

    def map_fields(
        self, source_fields: list[str], target_fields: list[str]
    ) -> MappingRequest:
        """
        Calls the OpenAI API to generate a draft mapping between two datasets.
        """
        field_mapping = self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": (
                        "You are an expert at structured data extraction. "
                        "You will be given two lists of phrases, one is the headers "
                        "for a target data file, and the other a set of descriptions "
                        "for columns of source data. "
                        "Match each target header to the best matching source "
                        "description, but match a header to None if a good match does "
                        "not exist. "
                        "Return the matched target headers and source descriptions using the provided structure."  # noqa
                    ),
                },
                {
                    "role": "user",
                    "content": (
                        f"These are the target headers: {target_fields}\n"
                        f"These are the source descriptions: {source_fields}"
                    ),
                },
            ],
            response_format=MappingRequest,
        )
        mappings = field_mapping.choices[0].message.parsed

        return mappings

    def map_values(
        self, values: list[tuple[str, set[str], list[str | None] | None]], language: str
    ) -> ValuesRequest:
        """
        Calls the OpenAI API to generate a set of value mappings for the fields.
        """
        value_mapping = self.client.beta.chat.completions.parse(
            model=self.model,
            messages=[
                {
                    "role": "system",
                    "content": (
                        "You are an expert at structured data extraction. "
                        "You will be given a list of tuples, where each tuple contains "
                        "three sets of string values. "
                        "The first set contains field names for a dataset."
                        "The second set contains values from a source dataset in "
                        f"{language}, and the third set contains target values for an "
                        "english-language transformed dataset. "
                        "Match all the values in the second set to the appropriate "
                        "values in the third set. "
                        "Return a list of dictionaries, where each dictionary contains "
                        "the field name as a key, and a dictionary containing "
                        "source values as keys, and the target text as values, "
                        "as the values. For example, the result should look like this: "
                        "[{'field_name_1': {'source_value_a': 'target_value_a', "
                        "'source_value_b': 'target_value_b'}, 'field_name_2':{...}]"
                        "using the provided structure."
                    ),
                },
                {
                    "role": "user",
                    "content": (
                        f"These are the field, source, target value sets: {values}"
                    ),
                },
            ],
            response_format=ValuesRequest,
        )
        mappings = value_mapping.choices[0].message.parsed

        return mappings



import argparse
import sys

try:
    from .dict_writer import DictWriter, create_dict, generate_descriptions
    from .make_toml import ParserGenerator, create_parser
    from .mapping import Mapper, create_mapping
except ImportError:  # pragma: no cover
    raise ImportError(
        "autoparser is not available. Import as 'adtl[autoparser]' to use."
    )

__all__ = [
    "DictWriter",
    "create_dict",
    "generate_descriptions",
    "Mapper",
    "create_mapping",
    "ParserGenerator",
    "create_parser",
]

from .dict_writer import main as make_dd_main
from .make_toml import main as make_toml_main
from .mapping import main as csv_mapping_main


def main():
    parser = argparse.ArgumentParser(
        description="adtl-autoparser: A tool for creating data dictionaries and parsers"
    )

    subparsers = parser.add_subparsers(title="subcommands", dest="subcommand")

    # Subcommand: create-dict
    parser_create_dict = subparsers.add_parser(
        "create-dict",
        help="Create a data dictionary from a dataset",
    )
    parser_create_dict.set_defaults(func=make_dd_main)

    # Subcommand: create-mapping
    parser_create_mapping = subparsers.add_parser(
        "create-mapping",
        help="Create initial CSV mapping from data dictionary (LLM key required)",
    )
    parser_create_mapping.set_defaults(func=csv_mapping_main)

    # Subcommand: create-parser
    parser_create_parser = subparsers.add_parser(
        "create-parser",
        help="Generate TOML parser from CSV mapping file",
    )
    parser_create_parser.set_defaults(func=make_toml_main)

    args, unknown_args = parser.parse_known_args()

    if args.subcommand is None:
        parser.print_help()
        sys.exit(1)

    # Call the appropriate function with remaining arguments
    args.func(unknown_args)
